#!/usr/bin/python3
#
# -*- coding: utf-8 -*-
#
# Copyright 2022 Univention GmbH
#
# http://www.univention.de/
#
# All rights reserved.
#
# The source code of this program is made available
# under the terms of the GNU Affero General Public License version 3
# (GNU AGPL V3) as published by the Free Software Foundation.
#
# Binary versions of this program provided by Univention to you as
# well as other copyrighted, protected or trademarked materials like
# Logos, graphics, fonts, specific documentations and configurations,
# cryptographic keys etc. are subject to a license agreement between
# you and Univention and not subject to the GNU AGPL V3.
#
# In the case you use this program under the terms of the GNU AGPL V3,
# the program is provided in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License with the Debian GNU/Linux or Univention distribution in file
# /usr/share/common-licenses/AGPL-3; if not, see
# <http://www.gnu.org/licenses/>.

import asyncio
import json
import logging
import multiprocessing
import os
import re
import subprocess
from argparse import ArgumentParser, ArgumentTypeError, RawDescriptionHelpFormatter
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
from os import listdir
from os.path import isfile, join
from typing import Awaitable, Callable, List, Tuple

import lazy_object_proxy
from ldap3 import Entry

from ucsschool_id_connector.ldap_access import LDAPAccess
from ucsschool_id_connector.models import SchoolAuthorityConfiguration
from ucsschool_id_connector.utils import ConsoleAndFileLogging

try:
    from verify import dn_to_id, ldap_object_to_school, ldap_object_to_school_class
except ImportError:
    import sys

    sys.path.append("/var/lib/univention-appcenter/apps/ucsschool-id-connector/conf/plugins/packages")
    from idbroker.verify import dn_to_id, ldap_object_to_school, ldap_object_to_school_class


try:
    # works in IDE, but not on the cli
    from .id_broker_client import (
        IDBrokerError,
        IDBrokerNotFoundError,
        IDBrokerObject,
        IDBrokerSchool,
        IDBrokerSchoolClass,
        ProvisioningAPIClient,
        School,
        SchoolClass,
    )
except ImportError:
    # works on the cli in ID Connector Docker container
    # make `idbroker` package importable
    import sys

    sys.path.append("/var/lib/univention-appcenter/apps/ucsschool-id-connector/conf/plugins/packages")

    from idbroker.id_broker_client import (
        IDBrokerError,
        IDBrokerNotFoundError,
        IDBrokerObject,
        IDBrokerSchool,
        IDBrokerSchoolClass,
        ProvisioningAPIClient,
        School,
        SchoolClass,
    )

SCHOOL_ATTRS = ["entryUUID", "ou", "displayName"]
SCHOOL_CLASS_ATTRS = ["entryUUID", "cn", "description", "uniqueMember"]
USER_ATTRS = ["entryUUID", "uid", "givenName", "sn", "memberOf", "ucsschoolRole"]
SCHOOL_CLASS_DN_PATTERN = re.compile(
    r"^cn=(?P<cn>.+?),cn=klassen,cn=schueler,cn=groups,ou=(?P<ou>.+?),dc=.+$"
)

MAX_THREADS = min(multiprocessing.cpu_count() * 2 + 1, 16)


logger: logging.Logger = lazy_object_proxy.Proxy(lambda: ConsoleAndFileLogging.get_logger(__name__))
ldap = LDAPAccess()

POOL_SIZE: int = None
auth_config: SchoolAuthorityConfiguration = None
SCHOOL_CLIENT: IDBrokerSchool = None
SCHOOLCLASS_CLIENT: IDBrokerSchoolClass = None
SYNC_MEMBERS: bool = None
DRY_MODE: bool = None


def list_dir(path: str) -> List:
    return [
        join(os.path.abspath(path), f)
        for f in listdir(os.path.abspath(path))
        if isfile(join(os.path.abspath(path), f))
    ]


async def sync_object(obj: IDBrokerObject) -> Tuple[str, bool, bool, bool]:
    global auth_config, DRY_MODE, SYNC_MEMBERS
    exists = False
    created = False
    updated = False
    obj_type: type = None
    name = ""

    if isinstance(obj, School):
        obj_type = School
        client = SCHOOL_CLIENT
    elif isinstance(obj, SchoolClass):
        obj_type = SchoolClass
        if SYNC_MEMBERS:
            if obj.members:
                obj.members = [await dn_to_id(member) for member in obj.members]
        else:
            obj.members = []
        client = SCHOOLCLASS_CLIENT
    else:
        try:
            raise ValueError(f"Could not identify type of given object {obj.name} (type= {type(obj)})")
        except Exception:
            raise ValueError(
                f"Could not identify type of given object {obj.user_name} (type= {type(obj)})"
            )

    if obj_type == SchoolClass:
        name = f"{obj.school}-{obj.name}"
    else:
        name = obj.name
    exists = await client.exists(obj.id)
    if exists:
        if obj_type != School:
            print(f"update {obj_type.__name__} {name}")
            if not DRY_MODE:
                await client.update(obj)
            updated = True
    else:
        print(f"create {obj_type.__name__} {name}")
        if not DRY_MODE:
            await client.create(obj)
        created = True

    return [name, exists, created, updated]


async def parallelize(funcs, args, custom_worker_pool):
    workers = POOL_SIZE
    if custom_worker_pool:
        workers = custom_worker_pool
    results = []
    with ThreadPoolExecutor(max_workers=workers) as executor:
        for func in funcs:
            for arg in args:
                future = executor.submit(func, arg)
                while True:
                    if future.done():
                        results.append(await future.result())
                        # maybe evaluate how long the clients have been running, to avoid timeout errors?
                        # then they dont need to get initialized on every object task
                        break
    return results


async def get_remote_object_and_compare(
    local_obj: IDBrokerObject,
    client: ProvisioningAPIClient,
    compare_func: Callable[[IDBrokerObject, IDBrokerObject], Awaitable[bool]],
) -> bool:
    if isinstance(local_obj, list):
        for entry in local_obj:
            try:
                await client.get(entry.id)
            except IDBrokerNotFoundError:
                return False
            except IDBrokerError:
                exit(-1)
            return True
    else:
        try:
            await client.get(local_obj.id)
        except IDBrokerNotFoundError:
            return False
        except IDBrokerError:
            exit(-1)
        return True


async def get_classes(
    ldap_filter: str,
) -> List[SchoolClass]:
    ldap_filter = f"(&(ucsschoolRole=school_class:school:{ldap_filter})(memberUid=*))"
    objects: List[Entry] = await ldap.search(filter_s=ldap_filter, attributes=SCHOOL_CLASS_ATTRS)
    return [ldap_object_to_school_class(o) for o in objects]


async def get_ous(
    ldap_filter: str,
) -> List[School]:
    ldap_filter = f"(&(objectClass=ucsschoolOrganizationalUnit)(ou={ldap_filter}))"
    objects: List[Entry] = await ldap.search(
        base=os.environ.get("ldap_base"), filter_s=ldap_filter, attributes=SCHOOL_ATTRS
    )
    return [ldap_object_to_school(o) for o in objects]


def flatten_list(xss: list) -> list:
    return [x for xs in xss for x in xs]


async def sync_complete(ldap_filter: str, dry_mode: bool) -> None:
    ldap_results = await parallelize(
        funcs=[get_ous, get_classes], args=[ldap_filter], custom_worker_pool=2
    )
    ldap_results = flatten_list(ldap_results)
    await parallelize([sync_object], ldap_results, None)


def setup(config_path: str, opt: List) -> None:
    global auth_config, SYNC_MEMBERS, SCHOOL_CLIENT, SCHOOLCLASS_CLIENT, DRY_MODE

    for f in list_dir(config_path):
        if f.split(".")[1] == "json":
            config = json.load(open(f))
            if "id_broker" in config["plugin_configs"]:
                try:
                    auth_config = SchoolAuthorityConfiguration(
                        name=config["name"],
                        url=config["url"],
                        plugin_configs=config["plugin_configs"],
                    )
                    break
                except Exception:
                    pass
    SCHOOL_CLIENT = IDBrokerSchool(school_authority=auth_config, plugin_name="id_broker")
    SCHOOLCLASS_CLIENT = IDBrokerSchoolClass(school_authority=auth_config, plugin_name="id_broker")
    SYNC_MEMBERS = opt.members
    DRY_MODE = opt.dry_mode


def thread_range():
    def checker(arg):
        if arg.isdigit():
            threads = int(arg)
        else:
            raise ArgumentTypeError(
                "Invalid argument type! The entered value is not type int or negative."
            )
        if threads < 1:
            raise ArgumentTypeError("At least one thread is needed to run.")
        if threads > MAX_THREADS:
            raise ArgumentTypeError(
                f"The maximum allowed value for concurrent threads is {MAX_THREADS} (Entered: {arg})"
            )
        return threads

    return checker


if __name__ == "__main__":

    parser = ArgumentParser(description=__doc__, formatter_class=RawDescriptionHelpFormatter)
    parser.add_argument(
        "-d",
        "--dry-mode",
        help="fakes complete sync with no commitment to the IDBroker and displays which objects would be created",
        action="store_true",
    )
    parser.add_argument(
        "-m",
        "--members",
        help="enables the sync of users",
        action="store_true",
    )
    parser.add_argument(
        "-s",
        "--single",
        help="only sync a single school with its classes (and optionally members if opt -m issued)",
        type=str,
    )
    parser.add_argument(
        "-t",
        "--threads",
        type=thread_range(),
        default=4,
        help=f"defines the number of threads that are used to run the specified task (min=1, max= {MAX_THREADS})",
    )
    parser.add_argument(
        "-c",
        "--config_path",
        default="/var/lib/univention-appcenter/apps/ucsschool-id-connector/conf/school_authorities",
        help="defines the path where the configuration for school_authorities can be found",
    )

    parser.epilog = """Examples:
    $ initial_sync.py -t T:int
        syncs all ous and classes without members (add -m to include members)

    $ initial_sync.py -s OU:str -t T:int
        syncs all classes of the ou specified by OU without members (add -m to include members)

    $ initial_sync.py -d -t T:int
        fake sync as preview for all ous and classes without members (add -m to include members)

    $ initial_sync.py -d -s OU:str -t T:int
        fake sync of singular ou specified by OU without members (add -m to include members)

    """

    opt = parser.parse_args()

    POOL_SIZE = int(opt.threads)
    logger.debug(f"STARTING {__file__} (improved first time sync) with arguments {opt}")
    start = datetime.now()
    print(f"Running with {POOL_SIZE} threads")

    setup(config_path=opt.config_path, opt=opt)
    ldap_filter = "*"
    if opt.single:
        ldap_filter = str(opt.single)
    try:
        subprocess.check_output(["/etc/init.d/ucsschool-id-connector", "stop"])
        asyncio.run(sync_complete(ldap_filter=ldap_filter, dry_mode=opt.dry_mode))
    finally:
        subprocess.check_output(["/etc/init.d/ucsschool-id-connector", "start"])

    end = datetime.now()
    duration = (end - start).total_seconds()
    print(f"Time taken: {duration} seconds")
    logger.debug(f"DONE {__file__} (improved first time sync). Processed in: {duration} seconds")
